@@ -1,12 +1,12 @@
 /***************************************
-  $Header: /home/amb/wwwoffle/RCS/misc.c 2.33 1999/05/10 17:32:39 amb Exp $
+  $Header: /home/amb/wwwoffle/RCS/misc.c 2.40 2000/02/20 09:44:01 amb Exp $
 
-  WWWOFFLE - World Wide Web Offline Explorer - Version 2.4d.
+  WWWOFFLE - World Wide Web Offline Explorer - Version 2.5d.
   Miscellaneous HTTP / HTML functions.
   ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1997,98 Andrew M. Bishop
+  This file Copyright 1997,98,99,2000 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -44,7 +44,7 @@ URL *SplitURL(char *url)
  URL *Url=(URL*)malloc(sizeof(URL));
  char *copyurl,*mallocurl=malloc(strlen(url)+2);
  int i=0,n=0;
- char *colon,*slash,*at,*temppath,root[2];
+ char *colon,*slash,*at,*ques,*temppath,root[2];
 
  copyurl=mallocurl;
  strcpy(copyurl,url);
@@ -73,7 +73,8 @@ URL *SplitURL(char *url)
     slash=strchr(copyurl,'/');
    }
  else if(colon && !isdigit(*(colon+1)) &&
-         (!at || (slash && at>slash)))  /* http:www.foo.com/... */
+         (!slash || colon<slash) &&
+         (!at || (slash && at>slash)))  /* http:www.foo.com/...[:@]... */
    {
     *colon=0;
     Url->proto=(char*)malloc(colon-copyurl+1);
@@ -101,7 +102,7 @@ URL *SplitURL(char *url)
 
  if(at && at>copyurl && (!slash || slash>at))
    {
-    if(colon && at>colon)               /* user:pass@www.foo.com... */
+    if(colon && at>colon)               /* user:pass@www.foo.com...[/]... */
       {
        *colon=0;
        Url->user=URLDecode(copyurl,0);
@@ -110,7 +111,7 @@ URL *SplitURL(char *url)
        Url->pass=URLDecode(copyurl,0);
        copyurl=at+1;
       }
-    else                                /* user@www.foo.com... */
+    else                               /* user@www.foo.com...[:/]... */
       {
        *at=0;
        Url->user=URLDecode(copyurl,0);
@@ -127,6 +128,19 @@ URL *SplitURL(char *url)
     Url->pass=NULL;
    }
 
+ /* Arguments */
+
+ ques=strchr(copyurl,'?');
+
+ if(ques)                       /* ../path?... */
+   {
+    *ques++=0;
+    Url->args=(char*)malloc(strlen(ques)+1);
+    strcpy(Url->args,ques);
+   }
+ else
+    Url->args=NULL;
+
  /* Hostname */
 
  if(*copyurl=='/')              /* /path/... (local) */
@@ -139,25 +153,12 @@ URL *SplitURL(char *url)
     Url->host=copyurl;
     Url->local=0;
 
-    if(slash)                   /* www.foo.com/... */
+    if(slash && (!ques || slash<ques)) /* www.foo.com/...[?]... */
        copyurl=slash;
-    else                        /* www.foo.com */
+    else                        /* www.foo.com[?]... */
       {root[0]='/';root[1]=0;copyurl=root;}
    }
 
- /* Arguments */
-
- Url->args=NULL;
-
- for(i=0;copyurl[i];i++)
-    if(copyurl[i]=='?')
-      {
-       copyurl[i]=0;
-       Url->args=(char*)malloc(strlen(copyurl+i+1)+1);
-       strcpy(Url->args,copyurl+i+1);
-       break;
-      }
-
  /* Pathname */
 
  temppath=URLDecode(copyurl,0);
@@ -245,6 +246,19 @@ URL *SplitURL(char *url)
  else
     Url->link=Url->name;
 
+ Url->dir=Url->host;
+
+#if defined(__CYGWIN__)
+ if(strchr(Url->host,':'))
+   {
+    char *colon=strchr(Url->host,':');
+    Url->dir=(char*)malloc(strlen(Url->host));
+    strcpy(Url->dir,Url->host);
+
+    Url->dir[colon-Url->host]='!';
+   }
+#endif
+
  /* end */
 
  free(mallocurl);
@@ -310,6 +324,9 @@ void FreeURL(URL *Url)
  if(Url->name!=Url->file)
     free(Url->file);
 
+ if(Url->dir!=Url->host)
+    free(Url->dir);
+
  free(Url->name);
 
  if(Url->proto) free(Url->proto);
@@ -358,7 +375,6 @@ char *LinkURL(URL *Url,char *link)
       {
        int j;
        char *path=(char*)malloc(strlen(Url->path)+strlen(link)+2);
-       char *match;
 
        strcpy(path,Url->path);
 
@@ -370,36 +386,7 @@ char *LinkURL(URL *Url,char *link)
        strcat(path,"/");
        strcat(path,link);
 
-       while((match=strstr(path,"/../")))
-         {
-          char *prev=match; match+=3;
-          while(prev>path && *--prev!='/');
-          while((*prev++=*match++));
-         }
-
-       while((match=strstr(path,"/./")))
-         {
-          char *prev=match; match+=2;
-          while((*prev++=*match++));
-         }
-
-       while((match=strstr(path,"//")))
-         {
-          char *prev=match; match++;
-          while((*prev++=*match++));
-         }
-
-       match=&path[strlen(path)-2];
-       if(match>=path && !strcmp(match,"/."))
-          *match=0;
-
-       match=&path[strlen(path)-3];
-       if(match>=path && !strcmp(match,"/.."))
-          if(match==path)
-             *match=0;
-          else
-             while(match>path && *--match!='/')
-                *match=0;
+       CanonicaliseName(path);
 
        new=(char*)malloc(strlen(Url->proto)+strlen(Url->host)+strlen(path)+8);
        sprintf(new,"%s://%s%s",Url->proto,Url->host,path);
@@ -411,6 +398,76 @@ char *LinkURL(URL *Url,char *link)
 }
 
 
+/*++++++++++++++++++++++++++++++++++++++
+  Canonicalise a file name by removing '/../', '/./' and '//' references.
+
+  char *CanonicaliseName Returns the argument modified.
+
+  char *name The original name
+  ++++++++++++++++++++++++++++++++++++++*/
+
+char *CanonicaliseName(char *name)
+{
+ char *match,*name2;
+
+ match=name;
+ while((match=strstr(match,"/./")) || !strncmp(match=name,"./",2))
+   {
+    char *prev=match, *next=match+2;
+    while((*prev++=*next++));
+   }
+
+ match=name;
+ while((match=strstr(match,"//")))
+   {
+    char *prev=match, *next=match+1;
+    while((*prev++=*next++));
+   }
+
+ match=name2=name;
+ while((match=strstr(match,"/../")))
+   {
+    char *prev=match, *next=match+4;
+    if((prev-name2)==2 && !strncmp(name2,"../",3))
+      {name2+=3;match++;continue;}
+    while(prev>name2 && *--prev!='/');
+    match=prev;
+    if(*prev=='/')prev++;
+    while((*prev++=*next++));
+   }
+
+ match=&name[strlen(name)-2];
+ if(match>=name && !strcmp(match,"/."))
+    *match=0;
+
+ match=&name[strlen(name)-3];
+ if(match>=name && !strcmp(match,"/.."))
+    if(match==name)
+       *++match=0;
+    else
+       while(match>name && *--match!='/')
+          *match=0;
+
+#if 0 /* as used in cxref */
+
+ match=&name[strlen(name)-1];
+ if(match>name && !strcmp(match,"/"))
+    *match=0;
+
+ if(!*name)
+    *name='.',*(name+1)=0;
+
+#else /* as used in wwwoffle */
+
+ if(!*name || !strncmp(name,"../",3))
+    *name='/',*(name+1)=0;
+
+#endif
+
+ return(name);
+}
+
+
 /*++++++++++++++++++++++++++++++++++++++
   Decode a string that has been UrlEncoded.
 
@@ -452,10 +509,51 @@ char *URLDecode(const char *str, int isform)
 }
 
 
+/*++++++++++++++++++++++++++++++++++++++
+  Decode a URL argument string that has been UrlEncoded.
+
+  char *URLDecodeArgs Returns a malloced copy of the decoded string.
+
+  const char *str The string to be decoded.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+char *URLDecodeArgs(const char *str)
+{
+ int i,j;
+ char *copy=(char*)malloc(strlen(str)+1);
+
+ for(i=0,j=0;str[i];i++)
+    if(str[i]=='%')
+      {
+       unsigned int val=0;
+       i++;
+       if(str[i]>='a') val=str[i]-'a'+10;
+       else if(str[i]>='A') val=str[i]-'A'+10;
+       else val=str[i]-'0';
+       val*=16;
+       i++;
+       if(str[i]>='a') val+=str[i]-'a'+10;
+       else if(str[i]>='A') val+=str[i]-'A'+10;
+       else val+=str[i]-'0';
+
+       if(val=='&' || val=='=')
+          copy[j++]=str[i];
+       else
+          copy[j++]=val;
+      }
+    else
+       copy[j++]=str[i];
+
+ copy[j]=0;
+
+ return(copy);
+}
+
+
 /*++++++++++++++++++++++++++++++++++++++
   Encode a string using the UrlEncode method as used with the POST method.
 
-  char *UrlEncode Returns a malloced copy of the encoded string.
+  char *URLEncode Returns a malloced copy of the encoded string.
 
   const char *str The string to be encoded.
   ++++++++++++++++++++++++++++++++++++++*/
@@ -468,7 +566,7 @@ char *URLEncode(const char *str)
  for(i=0,j=0;str[i];i++)
     if(isalpha(str[i]) ||
        isdigit(str[i]) ||
-       str[i]=='/' || str[i]=='?' || str[i]==':' || str[i]=='@' || str[i]=='=' ||
+       str[i]=='/' || str[i]=='?' || str[i]=='@' || str[i]=='=' ||
        str[i]=='$' || str[i]=='-' || str[i]=='_' || str[i]=='.' || str[i]=='+' || str[i]=='!' ||
        str[i]=='*' || str[i]=='(' || str[i]==')' || str[i]==',' || str[i]=='\'')
        copy[j++]=str[i];
@@ -497,7 +595,7 @@ char *URLEncode(const char *str)
 /*++++++++++++++++++++++++++++++++++++++
   Encode a string using the UrlEncode method as used for passwords in URLs.
 
-  char *UrlEncodePassword Returns a malloced copy of the encoded string.
+  char *URLEncodePassword Returns a malloced copy of the encoded string.
 
   const char *str The string to be encoded.
   ++++++++++++++++++++++++++++++++++++++*/
@@ -611,7 +709,7 @@ char *RFC822Date(long t,int utc)
          tim->tm_hour,
          tim->tm_min,
          tim->tm_sec,
-#if defined(__FreeBSD__) || defined(__NetBSD__)
+#if defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__)
          utc?"GMT":tim->tm_zone
 #elif defined(__CYGWIN__)
          utc?"GMT":_tzname[tim->tm_isdst>0]
